<?php

declare(strict_types=1);

namespace Rowlinson\Api;

use Rowlinson\Api\Collections\PaginatedCollection;
use Rowlinson\Api\Exceptions\NotFoundException;
use Rowlinson\Api\Exceptions\UnauthorisedException;
use Rowlinson\Api\Requests\User\ImpersonateRequest;
use Rowlinson\Api\Requests\User\InviteRequest;
use Rowlinson\Api\Requests\User\LoginRequest;
use Rowlinson\Api\Requests\User\NewPasswordRequest;
use Rowlinson\Api\Requests\User\RegisterRequest;
use Rowlinson\Api\Requests\User\ResetPasswordRequest;
use Rowlinson\Api\Requests\User\UpdateRequest;
use Rowlinson\Api\Requests\User\VerifyCredentialsRequest;
use Rowlinson\Api\Requests\User\VerifyRequest;
use Rowlinson\Api\Responses\User\GetInviteResponse;
use Rowlinson\Api\Responses\User\ListAllSingleResponse;
use Rowlinson\Api\Responses\User\LoginResponse;
use Rowlinson\Api\Responses\User\MeResponse;

class UserClient extends AbstractClient
{
    /**
     * Set the API key that the credentials use
     *
     * Only makes sense in UserClient as it deals with authentication and impersonation
     */
    public function setApiKey(string $apiKey): void
    {
        $this->credentials->setApiKey($apiKey);
    }

    /**
     * Log in with the supplied credentials
     *
     * If the login is successful, the credentials are updated with the returned API key
     */
    public function logIn(string $email, string $password, ?string $role = null): ?LoginResponse
    {
        $resp = null;

        try {
            $resp = $this->post('auth/login', new LoginRequest(
                $this->credentials->getClientKey(),
                $this->credentials->getClientSecret(),
                $email,
                $password,
                $role
            ));
        } catch (UnauthorisedException $e) {
            return null;
        }

        return $this->serializer->deserialize((string)$resp->getBody(), LoginResponse::class, 'json');
    }

    /**
     * Destroy the API key of the currently logged in user
     */
    public function logout(): void
    {
        $this->post('auth/logout');
    }

    /**
     * Verify a 2FA challenge
     */
    public function verify(string $challengeId, string $token): ?LoginResponse
    {
        $resp = null;

        try {
            $resp = $this->post('auth/verify', new VerifyRequest(
                $this->credentials->getClientKey(),
                $this->credentials->getClientSecret(),
                $challengeId,
                $token
            ));
        } catch (UnauthorisedException $e) {
            return null;
        }

        return $this->serializer->deserialize((string)$resp->getBody(), LoginResponse::class, 'json');
    }

    /**
     * Verify the credentials of the currently logged in user
     */
    public function verifyCredentials(string $email, string $password, ?string $role): bool
    {
        try {
            $this->post('auth/verify_credentials', new VerifyCredentialsRequest(
                $email,
                $password,
                $role
            ));

            return true;
        } catch (UnauthorisedException $e) {
        }

        return false;
    }

    /**
     * Request a password reset for the user identified by $email
     */
    public function resetPassword(string $email, string $url): void
    {
        $this->post('auth/reset_password', new ResetPasswordRequest(
            $this->credentials->getClientKey(),
            $this->credentials->getClientSecret(),
            $email,
            $url
        ));
    }

    /**
     * From a password reset request, set a new password for a user
     */
    public function newPassword(string $code, string $password): bool
    {
        try {
            $this->post('auth/new_password', new NewPasswordRequest(
                $this->credentials->getClientKey(),
                $this->credentials->getClientSecret(),
                $code,
                $password
            ));
        } catch (NotFoundException $e) {
            return false;
        }

        return true;
    }

    /**
     * Get the currently authenticated user
     */
    public function getCurrentUser(): MeResponse
    {
        $resp = $this->get('users/me');

        return $this->serializer->deserialize((string)$resp->getBody(), MeResponse::class, 'json');
    }

    /**
     * Update the currently authenticated user
     */
    public function updateCurrentUser(
        ?string $firstName,
        ?string $lastName,
        ?string $jobTitle,
        ?string $email,
        ?string $mobileNumber,
        ?string $landlineNumber,
        ?bool $enabled2FA,
        ?string $password
    ): void {
        $this->put('users/me', new UpdateRequest(
            $firstName,
            $lastName,
            $jobTitle,
            $email,
            $mobileNumber,
            $landlineNumber,
            $enabled2FA,
            $password
        ));
    }

    public function invite(
        string $firstName,
        string $lastName,
        ?string $jobTitle,
        string $email,
        string $url
    ): void {
        $this->post('users/invite', new InviteRequest(
            $firstName,
            $lastName,
            $jobTitle,
            $email,
            $url
        ));
    }

    /**
     * Get an invite by its code
     */
    public function getInvite(string $code): ?GetInviteResponse
    {
        $resp = null;

        try {
            $resp = $this->get('users/invite/' . $code);
        } catch (NotFoundException $e) {
            return null;
        }

        return $this->serializer->deserialize((string)$resp->getBody(), GetInviteResponse::class, 'json');
    }

    /**
     * Register a new user via invite code
     */
    public function register(string $code, string $password, ?string $phoneMobile): bool
    {
        try {
            $this->post('users/invite/' . $code, new RegisterRequest(
                $password,
                $phoneMobile
            ));
        } catch (NotFoundException $e) {
            return false;
        }

        return true;
    }

    /**
     * List all users in the system with optional filtering by name and email address
     *
     * @return PaginatedCollection<ListAllSingleResponse>
     */
    public function listAll(?string $name, ?string $email, int $page): PaginatedCollection
    {
        $response = $this->get('users?' . http_build_query(array_filter([
            'page' => $page > 0 ? $page : 1,
            'email' => ($email !== null) ? trim($email) : null,
            'name' => ($name !== null) ? trim($name) : null,
        ])));

        $users = $this->serializer->deserialize(
            (string)$response->getBody(),
            'array<' . ListAllSingleResponse::class . '>', // @phpstan-ignore-line
            'json'
        );

        return new PaginatedCollection(
            $users,
            $page,
            $response->getPaginationPerPage(),
            $response->getPaginationTotalItems()
        );
    }

    /**
     * Impersonate the user with the supplied email address
     */
    public function impersonate(string $email): LoginResponse
    {
        $body = new ImpersonateRequest($email);
        $response = $this->post('auth/impersonate', $body);

        return $this->serializer->deserialize((string)$response->getBody(), LoginResponse::class, 'json');
    }

    /**
     * Remove a user
     */
    public function remove(string $email): void
    {
        $this->delete('users/remove/' . $email);
    }

    /**
     * Resend n invite to an $email
     */
    public function resendInvite(string $email): void
    {
        $this->post('users/invite', new InviteRequest(
            '  ', // first name
            '  ', // last name
            null, //job title
            $email,
            '  ', // url
        ));
    }
}
